////////////////////////////////////////////////////////////////////////
//                                                                    //
// Time-stamp: "03.06.04 09:28 CSystemStdIOE.cpp klaus@wachtler.de"   //
//                                                                    //
////////////////////////////////////////////////////////////////////////
//                                                                    //
// Starten eines Kommandos mit system(), mit der Mglichkeit          //
// dessen stdin, stdout, und stderr durch eigene Funktionen           //
// zu leiten.                                                         //
//                                                                    //
////////////////////////////////////////////////////////////////////////
//                                                                    //
//           ///| |||\    || ||||    //                               //
//          ////| |||\\   || ||||   //                                //
//         ////|| ||||\\  || ||||  //                                 //
//        //// || |||| \\ || |||| //                                  //
//       ////==|| ||||  \\|| ||||//                                   //
//      ////   || ||||   \\| |||//                                    //
//     ////    || ||||    \| ||//                                     //
//                           |//                                      //
//         ||||    /|    //  //  ///| ||||=====//  ////=======        //
//         ||||   /||   //  //  ////| ||||    //  ////                //
//         ||||  //||  //  //  ////|| ||||   //  ////                 //
//         |||| // || //  //  //// || ||||  //  ////                  //
//         ||||//  ||//  //  ////==|| ||||=((  ((((=======            //
//         |||//   |//  //  ////   || ||||  \\  \\\\                  //
//         |||/    |/  //  ////    || ||||   \\  \\\\=======          //
//                                                                    //
////////////////////////////////////////////////////////////////////////
//                                                                    //
// Copyright:                                                         //
/////////////                                                         //
//                                                                    //
// (C) 2000-2004 AnyWare                                              //
//                                                                    //
// Dipl.Ing. Klaus Wachtler                                           //
// Breidingstr. 17                                                    //
// 29614 Soltau                                                       //
//                                                                    //
// Tel. (+49)-5191-70 2 71, (+49)-171/45 53 039                       //
// eMail AnyWare@Wachtler.de                                          //
//                                                                    //
// Nutzung erlaubt gem der GNU Public License (GPL);                //
// siehe www.fsf.org beispielsweise.                                  //
//                                                                    //
////////////////////////////////////////////////////////////////////////
//                                                                    //
// Siehe CSystemStdIOE.h!                                             //
/////////////////////////                                             //
//                                                                    //
////////////////////////////////////////////////////////////////////////
//                                                                    //
// Geschichte:                                                        //
//////////////                                                        //
//                                                                    //
// Datum:     Name: Vers. Aenderung:                                  //
// 30.04.2000 kw    0.0   erster Entwurf                              //
// 01.05.2000 kw          luft auch unter NT                         //
// 03.06.2004 kw          getestet mit gcc 3.2.3/Linux und VC6 sowie  //
//                        VC++ -NET 2002 unter W2k,                   //
//                        unter der GPL freigegeben                   //
//                                                                    //
////////////////////////////////////////////////////////////////////////


#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>

#include "CSystemStdIOE.h"

#ifdef _WIN32
#include <windows.h>
#endif /* ifdef _WIN32 */


// Die gestartete Task sollte keine Memberfunktion sein.
// Deshalb nehmen wir eine normale Funktion, und bergeben ihr
// alles, was sie in ihrem armseligen Leben braucht, in einer
// struct (stdioe_task_f fr die Userfunktionen) bzw.
// als String (cmd_task_f fr das eigentliche Kommando):

#ifdef _WIN32
static void __cdecl stdioe_task_f( void *startinfo_p )
#else /* ifdef _WIN32 */
static void * stdioe_task_f( void *startinfo_p )
#endif /* ifdef _WIN32 else ... */
{
  CSystemStdIOE::task_f_startinfo_t *si_p
    = (CSystemStdIOE::task_f_startinfo_t *)startinfo_p;

#ifdef DEBUGPRINT
  printf( "Task 0x%p( (int)%d ) gestartet\n",
          si_p->pf,
          *(si_p->fd)
          );
#endif // DEBUGPRINT

  // Benutzerfunktion starten:
  si_p->pf( *(si_p->fd) );

  // Datei schlieen:
  close( *(si_p->fd) );
  *(si_p->fd) = -1;

  // Vermerken, da wir fertig sind:
  *(si_p->isrunnnig) = false;

#ifdef DEBUGPRINT
  printf( "Task 0x%p( (int)%d ) beendet\n",
          si_p->pf,
          *(si_p->fd)
          );
#endif // DEBUGPRINT

  pthread_exit( NULL );

#ifdef _WIN32
  return;
#else /* ifdef _WIN32 */
  return NULL;
#endif /* ifdef _WIN32 else ... */

} // void *stdioe_task_f( void * )

#ifdef _WIN32
static void __cdecl cmd_task_f( void *startinfo_p )
#else /* ifdef _WIN32 */
static void * cmd_task_f( void *startinfo_p )
#endif /* ifdef _WIN32 else ... */
{
  CSystemStdIOE::task_f_cmd_startinfo_t *si_p
    = (CSystemStdIOE::task_f_cmd_startinfo_t *)startinfo_p;

  *(si_p->ret_p) = 0;
#ifdef DEBUGPRINT
  printf( "c: system( %s ) gestartet\n",
          si_p->cmd
          );
#endif // DEBUGPRINT

  int ret;
#ifdef DEBUGPRINT
  printf( "Jetzt kommt system( \"%s\" )\n", si_p->cmd );
#endif // DEBUGPRINT
  ret = system( si_p->cmd );
#ifdef DEBUGPRINT
  printf( "system( \"%s\" )=%d\n", si_p->cmd, ret );
#endif // DEBUGPRINT
  *(si_p->ret_p) = ret;
  //*(si_p->ret_p) = system( si_p->cmd );

  // Alle Dateien schlieen, soweit sie verwendet wurden:
  for( int ifd=0; ifd<3; ifd++ )
    {
      if( *(si_p->fds[ifd]) != -1 )
        {
          close( *(si_p->fds[ifd]) );
          *(si_p->fds[ifd]) = -1;
        }
    }

  // Vermerken, da wir fertig sind:
  *(si_p->isrunnnig) = false;

  //printf( "c: system( %s ) beendet\n",
  //        si_p->cmd
  //        );

  pthread_exit( NULL );

#ifdef _WIN32
  return;
#else /* ifdef _WIN32 */
  return NULL;
#endif /* ifdef _WIN32 else ... */

} // void *cmd_task_f( void * )

CSystemStdIOE::CSystemStdIOE( const char *Command        /* = NULL */,
                              userfd_pf_t User_Stdin_pf  /* = NULL */,
                              userfd_pf_t User_Stdout_pf /* = NULL */,
                              userfd_pf_t User_Stderr_pf /* = NULL */
                              )
  :  command( NULL ),
     user_stdin_pf( NULL ),
     user_stdout_pf( NULL ),
     user_stderr_pf( NULL ),
     stdin_task_started( false ),
     stdout_task_started( false ),
     stderr_task_started( false )
{
  ResetAlleFD();

  SetCommand( Command );
  SetStdin_f( User_Stdin_pf );
  SetStdout_f( User_Stdout_pf );
  SetStderr_f( User_Stderr_pf );

  retWert = 0;

} // CSystemStdIOE::CSystemStdIOE( const char *Command... )

CSystemStdIOE::~CSystemStdIOE()
{
  SetCommand();
} // CSystemStdIOE::~CSystemStdIOE( const char *command )

void CSystemStdIOE::ResetAlleFD()
{
  fdi[READENDE] = fdi[WRITEENDE] = -1;
  fdo[READENDE] = fdo[WRITEENDE] = -1;
  fde[READENDE] = fde[WRITEENDE] = -1;
} // void CSystemStdIOE::ResetAlleFD()

void CSystemStdIOE::CloseAlleFD()
{
  if( fdi[READENDE]!=-1 )
    {
      close( fdi[READENDE] );
    }
  if( fdi[WRITEENDE]!=-1 )
    {
      close( fdi[WRITEENDE] );
    }
  if( fdo[READENDE]!=-1 )
    {
      close( fdo[READENDE] );
    }
  if( fdo[WRITEENDE]!=-1 )
    {
      close( fdo[WRITEENDE] );
    }
  if( fde[READENDE]!=-1 )
    {
      close( fde[READENDE] );
    }
  if( fde[WRITEENDE]!=-1 )
    {
      close( fde[WRITEENDE] );
    }
  ResetAlleFD();
} // void CSystemStdIOE::CloseAlleFD()

void
CSystemStdIOE::SetCommand( const char *Command /* = NULL */ )
{
  if( command )
    {
      delete[] command;
      command = NULL;
    }

  if( Command )
    {
      command = new char[strlen( Command )+1];
      strcpy( command, Command );
    }

} // CSystemStdIOE::SetCommand( const char *Command /* = NULL */ ...

void
CSystemStdIOE::SetStdin_f( userfd_pf_t User_Stdin_pf /* = NULL */ )
{
  user_stdin_pf = User_Stdin_pf;
} // void SetStdin_f( userfd_pf_t User_Stdin_pf /* = NULL */ )

void
CSystemStdIOE::SetStdout_f( userfd_pf_t User_Stdout_pf /* = NULL */ )
{
  user_stdout_pf = User_Stdout_pf;
} // void SetStdout_f( userfd_pf_t User_Stdout_pf /* = NULL */ )

void
CSystemStdIOE::SetStderr_f( userfd_pf_t User_Stderr_pf /* = NULL */ )
{
  user_stderr_pf = User_Stderr_pf;
} // void SetStderr_f( userfd_pf_t User_Stderr_pf /* = NULL */ )


void CSystemStdIOE::Run( const char *Command /* = NULL */ )
{
  if( Command )
    {
      SetCommand( Command );
    }

  if( command )
    { // Kommando angegeben

      // Aus dem Kommando machen wir eine Kopie, an die ggf. die
      // Umlenkung fr stdin ("<&..."), stdout (">&..."), und
      // stderr ("2>&...") angehngt werden.
      // Weil vielleicht jemand so schusselig ist, schon in dem
      // Kommando etwas umzulenken, und trotzdem eine Funktion dafr
      // anzugeben, packen wir das ganze Kommando wie es ist in
      // Klammern, und setzen unsere Umlenkung dahinter.
      // Dann sieht die angegebene Funktion zwar nichts,
      // aber die shell, die das Ganze ausfhrt, mault auch nicht.

      char puffer[200];
      char *commandredir = NULL;

      try
        {
          // Platz fr das Kommando mit "<&..." und ">&..." und
          // "2>&..." schaffen:
          commandredir = new char[strlen(command)+100];
          strcpy( commandredir, "( " );
          strcat( commandredir, command );
          strcat( commandredir, " )" );

          // Ggf. Standardeingabe anhngen:
          if( user_stdin_pf )
            { // Funktion angegeben, die
              // eine Standardeingabe liefert:
              if( pipe( fdi ) )
                { // Fehler, keine pipe mglich!
                  fdi[READENDE] = fdi[WRITEENDE] = -1;
                  throw( CSystemStdIOEErr( "pipe error (stdin)" ) );
                }
              //printf( "fdi[]={%d,%d}\n", fdi[0], fdi[1] );
              snprintf( puffer,
                        sizeof(puffer),
                        " <&%d",
                        fdi[READENDE]
                        );
              strcat( commandredir, puffer );
            }

          // Ggf. Standardausgabe anhngen:
          if( user_stdout_pf )
            { // Funktion angegeben, die
              // die Standardausgabe vernascht:

              //// Unter Win32 kann man so leider nicht die
              //// Standardfehlerausgabe umleiten.
              //throw( CSystemStdIOEErr( "Umleitung von stderr unter"
              //                         " WIN32 leider nicht mglich!"
              //                         )
              //       );


              if( pipe( fdo ) )
                { // Fehler, keine pipe mglich!
                  fdo[READENDE] = fdo[WRITEENDE] = -1;
                  throw( CSystemStdIOEErr( "pipe error (stdout)" ) );
                }
              //printf( "fdo[]={%d,%d}\n", fdo[0], fdo[1] );
              snprintf( puffer,
                        sizeof(puffer),
                        " >&%d",
                        fdo[WRITEENDE]
                        );
              strcat( commandredir, puffer );
            }

          // Ggf. Standardfehlerausgabe anhngen:
          if( user_stderr_pf )
            { // Funktion angegeben, die
              // die Standardfehlerausgabe vernascht:
              if( pipe( fde ) )
                { // Fehler, keine pipe mglich!
                  fde[READENDE] = fde[WRITEENDE] = -1;
                  throw( CSystemStdIOEErr( "pipe error (stderr)" ) );
                }
#ifdef DEBUGPRINT
              printf( "fde[]={%d,%d}\n", fde[0], fde[1] );
#endif // DEBUGPRINT
              snprintf( puffer,
                        sizeof(puffer),
                        " 2>&%d",
                        fde[WRITEENDE]
                        );
              strcat( commandredir, puffer );
            }

          // jetzt starten wir fr jede angegebene Umleitung eine Task.
          // 1) stdin
          if( user_stdin_pf )
            {
              startinfo_i.pf = user_stdin_pf;
              startinfo_i.fd = &fdi[WRITEENDE];
              startinfo_i.isrunnnig = &stdin_task_started;

#ifdef DEBUGPRINT
              printf( "i: Task 0x%p( (int)%d ) starten...\n",
                      startinfo_i.pf,
                      *(startinfo_i.fd)
                      );
#endif // DEBUGPRINT
              stdin_task_started = true;
              if( pthread_create( &stdin_task_thread_id,
                                  NULL,
                                  stdioe_task_f,
                                  &startinfo_i
                                  )
                  )
                {
                  stdin_task_started = false;
                  throw( CSystemStdIOEErr( "thread create "
                                           "error (stdin)"
                                           )
                         );
                }
            }
          // 2) stdout
          if( user_stdout_pf )
            {
              startinfo_o.pf = user_stdout_pf;
              startinfo_o.fd = &fdo[READENDE];
              startinfo_o.isrunnnig = &stdout_task_started;

#ifdef DEBUGPRINT
              printf( "o: Task 0x%p( (int)%d ) starten...\n",
                      startinfo_o.pf,
                      *(startinfo_o.fd)
                      );
#endif // DEBUGPRINT
              stdout_task_started = true;
              if( pthread_create( &stdout_task_thread_id,
                                  NULL,
                                  stdioe_task_f,
                                  &startinfo_o
                                  )
                  )
                {
                  stdout_task_started = false;
                  throw( CSystemStdIOEErr( "thread create "
                                           "error (stdout)"
                                           )
                         );
                }
            }
          // 3) stderr
          if( user_stderr_pf )
            {
              startinfo_e.pf = user_stderr_pf;
              startinfo_e.fd = &fde[READENDE];
              startinfo_e.isrunnnig = &stderr_task_started;

#ifdef DEBUGPRINT
              printf( "e: Task 0x%p( (int)%d ) starten...\n",
                      startinfo_e.pf,
                      *(startinfo_e.fd)
                      );
#endif // DEBUGPRINT
              stderr_task_started = true;
              if( pthread_create( &stderr_task_thread_id,
                                  NULL,
                                  stdioe_task_f,
                                  &startinfo_e
                                  )
                  )
                {
                  stderr_task_started = false;
                  throw( CSystemStdIOEErr( "thread create "
                                           "error (stderr)"
                                           )
                         );
                }
            }

          // vierte und letzte Task fr das eigentliche Kommando:
          cmd_startinfo.cmd = commandredir;
          cmd_startinfo.fds[0] = &fdi[READENDE];
          cmd_startinfo.fds[1] = &fdo[WRITEENDE];
          cmd_startinfo.fds[2] = &fde[WRITEENDE];
          cmd_startinfo.isrunnnig = &cmd_task_started;
          cmd_startinfo.ret_p = &retWert;
#ifdef DEBUGPRINT
          printf( "Kommando=\"%s\"\n", commandredir );
#endif // DEBUGPRINT
          cmd_task_started = true;
          if( pthread_create( &cmd_task_thread_id,
                              NULL,
                              cmd_task_f,
                              &cmd_startinfo
                              )
              )
            {
              cmd_task_started = false;
              throw( CSystemStdIOEErr( "thread create "
                                       "error (cmd)"
                                       )
                     );
            }
        }
      catch( ... )
        {
          // Aufrumen

          if( stdin_task_started )
            {

            }
          if( commandredir )
            {
              delete[] commandredir;
              commandredir = NULL;
            }
          CloseAlleFD();

          throw;
        }

      if( commandredir )
      {
        delete[] commandredir;
        commandredir = NULL;
      }
    } // Kommando angegeben

} // CSystemStdIOE::Run( const char *command /* = NULL */ )

int CSystemStdIOE::Wait()
{
#ifdef _WIN32

  while( isCommandRunning()
         ||
         isStdinRunning()
         ||
         isStdoutRunning()
         ||
         isStderrRunning()
         )
    {
      Sleep( 10 ); // 10 msec schlafen
    }

#else /* ifdef _WIN32 */

  if( isCommandRunning() )
    {
      pthread_join( cmd_task_thread_id, NULL );
    }

  if( isStdinRunning() )
    {
      pthread_join( stdin_task_thread_id, NULL );
    }

  if( isStdoutRunning() )
    {
      pthread_join( stdout_task_thread_id, NULL );
    }

  if( isStderrRunning() )
    {
      pthread_join( stderr_task_thread_id, NULL );
    }

#endif /* ifdef _WIN32 else ... */

  return retWert;
} // CSystemStdIOE::int Wait()

//
//////////////////////////////////////////////////////////////////////
