///////////////////////////////////////////////////-*-c++-*-////////////
//                                                                    //
// Time-stamp: "03.06.04 09:23 CSystemStdIOE.h klaus@wachtler.de"     //
//                                                                    //
////////////////////////////////////////////////////////////////////////
//                                                                    //
// Starten eines Kommandos mit system(), mit der Mglichkeit          //
// dessen stdin, stdout, und stderr durch eigene Funktionen           //
// zu leiten.                                                         //
//                                                                    //
////////////////////////////////////////////////////////////////////////
//                                                                    //
//           ///| |||\    || ||||    //                               //
//          ////| |||\\   || ||||   //                                //
//         ////|| ||||\\  || ||||  //                                 //
//        //// || |||| \\ || |||| //                                  //
//       ////==|| ||||  \\|| ||||//                                   //
//      ////   || ||||   \\| |||//                                    //
//     ////    || ||||    \| ||//                                     //
//                           |//                                      //
//         ||||    /|    //  //  ///| ||||=====//  ////=======        //
//         ||||   /||   //  //  ////| ||||    //  ////                //
//         ||||  //||  //  //  ////|| ||||   //  ////                 //
//         |||| // || //  //  //// || ||||  //  ////                  //
//         ||||//  ||//  //  ////==|| ||||=((  ((((=======            //
//         |||//   |//  //  ////   || ||||  \\  \\\\                  //
//         |||/    |/  //  ////    || ||||   \\  \\\\=======          //
//                                                                    //
////////////////////////////////////////////////////////////////////////
//                                                                    //
// Copyright:                                                         //
/////////////                                                         //
//                                                                    //
// (C) 2000-2004 AnyWare                                              //
//                                                                    //
// Dipl.Ing. Klaus Wachtler                                           //
// Breidingstr. 17                                                    //
// 29614 Soltau                                                       //
//                                                                    //
// Tel. (+49)-5191-70 2 71, (+49)-171/45 53 039                       //
// eMail AnyWare@Wachtler.de                                          //
//                                                                    //
// Nutzung erlaubt gem der GNU Public License (GPL);                //
// siehe www.fsf.org beispielsweise.                                  //
//                                                                    //
////////////////////////////////////////////////////////////////////////
//                                                                    //
// Aufgabenstellung:                                                  //
////////////////////                                                  //
//                                                                    //
// Mithilfe dieser Klasse kann man irgendwelche Programme laufen      //
// lassen, und dabei ihre Standardeingabe, Standardausgabe, und       //
// Standardfehlerausgabe durch eigene Funktionen erzeugen bzw.        //
// abfangen.                                                          //
//                                                                    //
////////////////////////////////////////////////////////////////////////
//                                                                    //
// Verwendung:                                                        //
//////////////                                                        //
//                                                                    //
// Es gibt drei Funktionen, die der Aufrufer definieren kann:         //
// 1) eine Funktion, um die Standardeingabe fr das Kommando          //
//    zu liefern,                                                     //
// 2) eine um die Standardausgabe abzufangen, und                     //
// 3) eine um die Standardfehlerausgabe abzufangen.                   //
// Jede dieser Funktionen erhlt einen file descriptor (Typ int),     //
// und gibt nichts zurck (Rckgabetyp: void).                        //
// Die Funktion, die die Standardeingabe liefert, mu zu diesem       //
// file descriptor schreiben (write unter Linux, _write unter W32);   //
// diese Ausgabe geht als Standardeingabe an das Kommando.            //
// Die Funktionen, die die Standardausgabe und die                    //
// Standardfehlerausgabe abfangen, bekommen ber ihren                //
// file descriptor die Ausgabe des Kommandos als ihre                 //
// eigene Eingabe (zu lesen mit read oder _read).                     //
// Wenn die Funktionen fertig sind, brauchen sie die Dateien          //
// nicht zu schlieen, das macht die Klasse selbst.                   //
// Die Funktion fr die Standardeingabe                               //
// sollte sich beenden, wenn sie alles geschrieben hat.               //
// Die beiden Funktionen, die eine Ausgabe abfangen,                  //
// sollten sich beenden, wenn ein Lesefehler auftritt                 //
// (bzw. EOF gelesen wird).                                           //
//                                                                    //
// Der Aufrufer definiert sich ein Objekt vom Typ CSystemStdIOE.      //
// Das auszufhrende Kommando und maximal drei Zeiger auf die         //
// Funktionen knnen entweder schon dem Konstruktor bergeben werden, //
// oder mit den Funktionen SetCommand(), SetStdin_f(), SetStdout_f()  //
// und SetStderr_f(), oder sptestens beim Aufruf von Run() gesetzt   //
// werden.                                                            //
//                                                                    //
// Mit Run() wird das Kommando gestartet. Dabei laufen die (maximal)  //
// drei Funktionen und das Kommando im Hintergrund; Run() kehrt       //
// direkt nach dem Start wieder zurck.                               //
//                                                                    //
// Wenn der Aufrufer auf das Ende des Kommandos warten will, dann     //
// kann er nach Run() ein Wait() aufrufen. Dieses liefert als         //
// Rckgabewert die Rckgabe des ausgefhrten Kommandos.              //
// Geliefert wird der Wert wie ich ihn von system() bekomme.          //
// Laut man system ist das der Rckgabewert des aufgerufenen          //
// Kommandos. Es scheint aber so zu sein, da dieser Wert um          //
// 8 Bit nach links geschoben ist, bzw. mit 256 multipliziert.        //
// Bei einem exit(2) im aufgerufenen Kommando liefert Wait()          //
// also den Wert 512 (unter Linux).                                   //
//                                                                    //
// Wenn der Aufrufer aber in der Zwischenzeit etwas Besseres zu tun   //
// hat, dann kann er das machen, und gelegentlich mit                 //
// isStdinRunning(), isStdoutRunning(), isStderrRunning(),            //
// und isCommandRunning() kontrollieren, ob die einzelnen             //
// Tasks noch laufen.                                                 //
// Wahrscheinlich ist isRunning() das Sinnvolle, es liefert true,     //
// wenn irgendeine der Tasks luft.                                   //
// Den Rckgabewert kann man dann wieder mit Wait() abfragen.         //
//                                                                    //
// Theoretisch ist es auch mglich, whrend ein Run() luft, mit      //
// einem weiteren schon das nchste Kommando zu starten.              //
// Dann kann man aber die Rckgabewerte nicht mehr auseinanderhalten. //
// Auch die is...Running()-Funktionen liefern dann Unfug.             //
// Will man die Rckgabewerte haben, und trotzdem mehrere Sachen      //
// gleichzeitig laufen lassen, definiert man sich halt einfach        //
// mehrere CSystemStdIOE-Objekte.                                     //
//                                                                    //
// Bei Fehlern (pipe kann nicht geffnet werden etc.), wird eine      //
// Ausnahme vom Typ CSystemStdIOEErr geworfen (abgeleitet von         //
// exception).                                                        //
//                                                                    //
////////////////////////////////////////////////////////////////////////
//                                                                    //
// Geschichte:                                                        //
//////////////                                                        //
//                                                                    //
// Datum:     Name: Vers. Aenderung:                                  //
// 01.05.2000 kw    0.0   erster Entwurf, Doku                        //
// 03.06.2004 kw          getestet mit gcc 3.2.3/Linux und VC6 sowie  //
//                        VC++ -NET 2003 unter W2k,                   //
//                        unter der GPL freigegeben                   //
//                                                                    //
////////////////////////////////////////////////////////////////////////
//                                                                    //
// Benoetigt:                                                         //
/////////////                                                         //
//                                                                    //
// POSIX-threads (pthread) oder LIBCMT.LIB/MSVCRT.LIB                 //
//                                                                    //
////////////////////////////////////////////////////////////////////////
//                                                                    //
// Umgebung:                                                          //
////////////                                                          //
//                                                                    //
// ANSI C++ + Posix                                                   //
// Linux oder NT/W2k                                                  //
// (zustzlich zu POSIX sind int-Filedeskriptoren ntig)              //
//                                                                    //
////////////////////////////////////////////////////////////////////////
//                                                                    //
// Bemerkungen fr WIN32:                                             //
/////////////////////////                                             //
//                                                                    //
// Es ist eine multithreadfhige Version der Laufzeitbibliothek       //
// ntig, also LIBCMT.LIB oder MSVCRT.LIB.                            //
// Kompilieren mit /MT /GX                                            //
//                                                                    //
// Die Umleitung von stderr klappt unter Win32 eher nicht.            //
//                                                                    //
////////////////////////////////////////////////////////////////////////

#ifndef _CSystemStdIOE_H_
#define _CSystemStdIOE_H_

#ifdef _WIN32

#include <process.h>

// Umsetzung von Posix-Funktionen auf Win32-Funktionen:

// thread starten ist fast gleich; attr_p brauchen wir eh nicht.
// pthread_create() schreibt die thread id dahin, wo der erste Paramter
// hinzeigt; Win32-_beginthread() liefert die id dagegen als
// Rckgabewert (oder -1 im Fehlerfall).
#define pthread_create(id_p,attr_p,func_p,arg_p) \
        ( ( *id_p =_beginthread( func_p, 0, arg_p ) ), *id_p==-1 )

// Thrad beenden ist auch fast gleich; Rckgabewert egal:
#define pthread_exit(ststus) _endthread()

// so etwas wie pthread_join() gibt es in Wondows nicht.
// stattdessen mssen wir halt in Wait() zu Fu warten.
#define pthread_join(tid,status_p) (void)

// thread id ist in Win32 eine unsigned long:
typedef unsigned long pthread_t;

#include <io.h>
#include <fcntl.h>


// nur damit es anders ist, heit write() bei MS _write() etc.:
#define open(f,a) _open( f, a )
#define close(fd) _close( fd )
#define pipe(fds) _pipe( fds, 1024, _O_TEXT )
#define snprintf _snprintf

#else /* ifdef _WIN32 */

#include <unistd.h>
#include <pthread.h>

#endif /* ifdef _WIN32 else ... */

#include <stdexcept>
#include <string>


// Gelegentlich geworfene Ausnahme:
class CSystemStdIOEErr
  : public std::runtime_error
{
public:
      CSystemStdIOEErr( const std::string &_what = "CSystemStdIOEErr" )
        : runtime_error( std::string( "Fehler mit CSystemStdIOEErr\n" )
                         +
                         _what
                         )
  {}
};


//// CSystemStdIOEErr verwandelt einen (char*) in etwas, was bei Fehlern
//// ausgeworfen werden soll.
//// Wenn man nicht vorher etwas anderes definiert, dann wird bei
//// Fehlern ein (char*) ausgeworfen.
//// Will man stattdessen einen eigenen Fehlertyp ausgeworfen haben,
//// beispielsweise weil man sich eine Klasse CError definiert hat,
//// dann definiert sich einfach eine Umwandlung eines (char*) in diese
//// Klasse.
//// Wenn beispielsweise CError einen Konstruktor hat, der mit einem
//// (char*) initialisiert werden kann, dann definiert man sich
////   #define CSystemStdIOEErr(a) CError(a)
//// und schon werden nicht mehr (char*), sondern (CError) ausgeworfen.
//
//#ifndef CSystemStdIOEErr
//#define CSystemStdIOEErr(a) ((const char*)a)
//#endif // ifdef CSystemStdIOEErr

class CSystemStdIOE
{
private:

  // siehe: man 2 pipe
  enum { READENDE = 0, WRITEENDE = 1 };

  char *command;

  // Pipe fr die Standardeingabe des Kommandos:
  int fdi[2];
  // Pipe fr die Standardausgabe des Kommandos:
  int fdo[2];
  // Pipe fr die Standardfehlerausgabe des Kommandos:
  int fde[2];

  void ResetAlleFD();
  void CloseAlleFD();
public:

  // So mssen Funktionen aussehen, die
  // fr das Kommando Standardeingabe
  // produzieren, oder die Standardausgabe aufnehmen,
  // oder die Standardfehlerausgabe abfangen:
  typedef void userfd_f_t( int fd );
  // userfd_pf_t ist ein Zeiger auf eine solche Funktion:
  typedef userfd_f_t * userfd_pf_t;

  // Die drei Tasks fr i/o/e bekommen je eine solche struct
  // mitgegeben.
  // Sie finden hier die zu startende Userfunktion und
  // einen Zeiger auf das betreffende Dateihandle und
  // auf eine bool, die angibt, ob die Task luft.
  // Die Task, die das Kommando ausfhrt, findet hier
  // den Aufruf fr system() und maximal drei zu  schlieende
  // Dateihandles (oder -1).
  typedef struct
  {
    const char  *cmd;        // zu startendes Kommando
    int         *fds[3];     // umgelenkte fd's dafr
    bool        *isrunnnig;  // luft die Task?
    int         *ret_p;      // Rckgabewert
  } task_f_cmd_startinfo_t;

  typedef struct
  {
    userfd_pf_t  pf;        // zu startende User-Funktion
    int         *fd;        // file descriptor dafr
    bool        *isrunnnig; // luft die Task?
  } task_f_startinfo_t;

private:

  task_f_startinfo_t startinfo_i;
  task_f_startinfo_t startinfo_o;
  task_f_startinfo_t startinfo_e;

  task_f_cmd_startinfo_t cmd_startinfo;

  userfd_pf_t user_stdin_pf;
  userfd_pf_t user_stdout_pf;
  userfd_pf_t user_stderr_pf;

  // Es gibt je eine interne Funktion, die fr
  // stdin, stdout, und stderr als eigene
  // Task gestartet wird.
  // Jede dieser Funktionen ruft eine der drei
  // user-Funktionen auf.

  pthread_t stdin_task_thread_id;
  bool stdin_task_started;

  pthread_t stdout_task_thread_id;
  bool stdout_task_started;

  pthread_t stderr_task_thread_id;
  bool stderr_task_started;

  // Eine weitere Task lt das eigentliche
  // Kommando laufen:
  pthread_t cmd_task_thread_id;
  bool cmd_task_started;

  // Platz fr Rckgabewert:
  int   retWert;

public:

  CSystemStdIOE( const char *command        = NULL,
                 userfd_pf_t User_Stdin_pf  = NULL,
                 userfd_pf_t User_Stdout_pf = NULL,
                 userfd_pf_t User_Stderr_pf = NULL
                 );

  ~CSystemStdIOE();

  void SetCommand( const char *command = NULL );
  void SetStdin_f( userfd_pf_t User_Stdin_pf  = NULL );
  void SetStdout_f( userfd_pf_t User_Stdout_pf  = NULL );
  void SetStderr_f( userfd_pf_t User_Stderr_pf  = NULL );

  bool isStdinRunning() { return stdin_task_started; };
  bool isStdoutRunning() { return stdout_task_started; };
  bool isStderrRunning() { return stderr_task_started; };
  bool isCommandRunning() { return cmd_task_started; };
  bool isRunning()
    {
      return
        stdin_task_started
        ||
        stdout_task_started
        ||
        stderr_task_started
        ||
        cmd_task_started;
    };

  void Run( const char *command = NULL );
  int Wait();

}; // class CSystemStdIOE

#endif // _CSystemStdIOE_H_

//
//////////////////////////////////////////////////////////////////////

