//  Time-stamp: "03.06.04 09:23 tstSystemStdIOE.cpp klaus@wachtler.de"
//
// Testprogramm fr CSystemStdIOE.
// Es wird ein weiteres Programm (sinus.cpp) bentigt.
// Dieses wird in main() mit CSystemStdIOE::Run() gestartet.
// Dabei wird die Standardeingabe von sinus von der
// Funktion void liefere_stdin( int fd ) geliefert.
// Standardausgabe und Standardfehlerausgabe werden von
// void vernasche_stdout( int fd ) und
// void vernasche_stderr( int fd ) abgefangen.
//
// Verwendung unter Linux:
//   g++ -Wall sinus.cpp -o sinus
//   g++ -Wall tstSystemStdIOE.cpp CSystemStdIOE.cpp -o tstSystemStdIOE -lpthread
//   ./tstSystemStdIOE
//
// Verwendung unter Win32/VC++ (getestet: VC++ 6.0 und VC++ .NET 2003
// unter W2k):
//   cl sinus.cpp
//   cl /MT /GX tstSystemStdIOE.cpp CSystemStdIOE.cpp
//   tstSystemStdIOE
//
//
// Copyright:
//
// (C) 2000-2004 AnyWare
//
// Dipl.Ing. Klaus Wachtler
// Breidingstr. 17
// 29614 Soltau
//
// Tel. (+49)-5191-70 2 71, (+49)-171/45 53 039
// eMail AnyWare@Wachtler.de
//
// Nutzung erlaubt gem der GNU Public License (GPL);
// siehe www.fsf.org beispielsweise.


// auszufhrende Kommandozeile:
#ifdef _WIN32
#define SINUS_EXE  "sinus.exe"
#else
#define SINUS_EXE  "./sinus"
#endif


#include <string.h>
#include <stdio.h>
#include "CSystemStdIOE.h"

#ifdef _WIN32

#include <io.h>

// nur damit es anders ist, heit write() bei MS _write() etc.:
#define write(fd,p,l) _write( fd, p, l )
#define read(fd,p,l) _read( fd, p, l )

#endif /* ifdef _WIN32 else ... */

void liefere_stdin( int fd )
{
  for( double d=0.0; d<6.2; d+=0.1 )
    {
      char puffer[100];
      sprintf( puffer, "%f\n", d );
      write( fd, puffer, strlen(puffer) );
    }
  write( fd, "(end)\n", strlen("(end)\n") );

} // void liefere_stdin( int fd )

void vernasche_stdout( int fd )
{
  char zeile[1024];
  int spalte = 0;

  while( read( fd, zeile+spalte, 1 )==1 )
    {
      if( spalte==sizeof(zeile)-2 )
        {
          zeile[spalte+1] = '\0';
          printf( "vernasche_stdout(): <%s>\n", zeile );
          spalte = 0;
        }
      else if( zeile[spalte]=='\n' )
        {
          zeile[spalte] = '\0';
          printf( "vernasche_stdout(): <%s>\n", zeile );
          spalte = 0;
        }
      else
        {
          spalte++;
        }
    }

  if( spalte )
    {
      zeile[spalte+1] = '\0';
      printf( "vernasche_stdout(): <%s>\n", zeile );
    }

} // void vernasche_stdout( int fd )

void vernasche_stderr( int fd )
{
  char zeile[1024];
  int spalte = 0;

  while( read( fd, zeile+spalte, 1 )==1 )
    {
      if( spalte==sizeof(zeile)-2 )
        {
          zeile[spalte+1] = '\0';
          printf( "vernasche_stderr(): <%s>\n", zeile );
          spalte = 0;
        }
      else if( zeile[spalte]=='\n' )
        {
          zeile[spalte] = '\0';
          printf( "vernasche_stderr(): <%s>\n", zeile );
          spalte = 0;
        }
      else
        {
          spalte++;
        }
    }

  if( spalte )
    {
      zeile[spalte+1] = '\0';
      printf( "vernasche_stderr(): <%s>\n", zeile );
    }

} // void vernasche_stderr( int fd )

int main( int nargs, char **args )
{
  try
    {
      CSystemStdIOE  cmd;

      cmd.SetStdin_f( liefere_stdin );
      cmd.SetStdout_f( vernasche_stdout );
#ifndef _WIN32
      cmd.SetStderr_f( vernasche_stderr );
#endif

      cmd.Run( SINUS_EXE );

      printf( "isStdinRunning()   = %s\n",
              cmd.isStdinRunning() ? "y" : "n"
              );
      printf( "isStdoutRunning()  = %s\n",
              cmd.isStdoutRunning() ? "y" : "n"
              );
      printf( "isStderrRunning()  = %s\n",
              cmd.isStderrRunning() ? "y" : "n"
              );
      printf( "isCommandRunning() = %s\n",
              cmd.isCommandRunning() ? "y" : "n"
              );

      cmd.Wait();

      printf( "isStdinRunning()   = %s\n",
              cmd.isStdinRunning() ? "y" : "n"
              );
      printf( "isStdoutRunning()  = %s\n",
              cmd.isStdoutRunning() ? "y" : "n"
              );
      printf( "isStderrRunning()  = %s\n",
              cmd.isStderrRunning() ? "y" : "n"
              );
      printf( "isCommandRunning() = %s\n",
              cmd.isCommandRunning() ? "y" : "n"
              );

      cmd.Run( SINUS_EXE );
      int ret = cmd.Wait();

      printf( "ret = %d\n", ret );
    }
  catch( const char *s )
    {
      fprintf( stderr, "Fehler \"%s\" gefangen!\n", s );
    }
  catch( ... )
    {
      fprintf( stderr, "Unbekannten Fehler gefangen!\n" );
    }

  printf( "und Tsch!\n" );
  return 0;

} /* main( int nargs, char **args ) */
